import threading

from snmp.dispatcher import *
from snmp.manager.v2c import *
from snmp.manager.v3 import *
from snmp.message import *
import snmp.message.v2c
import snmp.message.v3
from snmp.security import *
from snmp.security.levels import *
from snmp.security.usm import *
from snmp.transport import *
from snmp.transport.udp import *

class DiscoveryGuard:
    def __init__(self):
        self.namespace = None
        self.refCount = 0

    def claim(self, namespace):
        acquired = True
        initialized = True

        if namespace != self.namespace:
            if self.refCount:
                acquired = False
            else:
                self.namespace = namespace
                initialized = False

        if acquired:
            self.refCount += 1
            #print(f"claim  (\"{self.namespace}\"): refCount = {self.refCount}")

        return acquired, initialized

    def release(self, namespace):
        assert self.namespace == namespace
        assert self.refCount > 0
        self.refCount -= 1
        #print(f"release(\"{self.namespace}\"): refCount = {self.refCount}")
        return self.refCount == 0

class UserEntry:
    def __init__(self, defaultSecurityLevel, credentials):
        self.credentials = credentials
        self.defaultSecurityLevel = defaultSecurityLevel

class NameSpace:
    def __init__(self, defaultUserName):
        self.defaultUserName = defaultUserName
        self.users = {}

    def __iter__(self):
        return self.users.items().__iter__()

    def __contains__(self, key):
        return self.users.__contains__(key)

    def addUser(self, userName, *args, **kwargs):
        self.users[userName] = UserEntry(*args, **kwargs)

    def getUser(self, userName):
        return self.users[userName]

class Engine:
    TRANSPORTS = {
        cls.DOMAIN: cls for cls in [
            UdpTransport,
        ]
    }

    UNSUPPORTED = "{} is not supported at this time"

    def __init__(self, lockType=threading.Lock,
            defaultDomain=TransportDomain.UDP,
            defaultVersion=MessageProcessingModel.SNMPv3,
            defaultSecurityModel=SecurityModel.USM,
            autowait=True):

        self.defaultDomain = defaultDomain
        self.defaultVersion = defaultVersion
        self.defaultSecurityModel = defaultSecurityModel
        self.autowaitDefault = autowait
        self.dispatcher = Dispatcher(lockType=lockType)

        self.lock = lockType()
        self.lockType = lockType

        self.engines = {}
        self.namespaces = {}

        self.transports = set()
        self.mpv2c = None
        self.mpv3 = None
        self.usm = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.shutdown()

    def shutdown(self):
        self.dispatcher.shutdown()

    def registerRemoteEngine(self, engineID, namespace):
        with self.lock:
            try:
                guard = self.engines[engineID]
            except KeyError:
                guard = DiscoveryGuard()
                self.engines[engineID] = guard

            acquired, initialized = guard.claim(namespace)
            if acquired and not initialized:
                space = self.namespaces[namespace]
                for userName, userEntry in space:
                    kwargs = self.localize(engineID, **userEntry.credentials)
                    self.usm.addUser(engineID, userName, **kwargs)

            return acquired

    def unregisterRemoteEngine(self, engineID, namespace):
        with self.lock:
            try:
                guard = self.engines[engineID]
            except KeyError:
                assert False, f"Engine {engineID} was never registered"
            else:
                if guard.release(namespace):
                    del self.engines[engineID]
                    #print(list(self.engines.keys()))

    @staticmethod
    def localize(engineID, authProtocol=None, authSecret=None,
                           privProtocol=None, privSecret=None):
        kwargs = dict()
        if authProtocol is not None:
            kwargs["authProtocol"] = authProtocol
            kwargs["authKey"] = authProtocol.localize(authSecret, engineID)

            if privProtocol is not None:
                kwargs["privProtocol"] = privProtocol
                kwargs["privKey"] = authProtocol.localize(privSecret, engineID)

        return kwargs

    def addUser(self, userName, authProtocol=None, authSecret=None,
            privProtocol=None, privSecret=None, secret=b"",
            default=False, defaultSecurityLevel=None, namespace=""):
        kwargs = dict()
        if authProtocol is None:
            maxSecurityLevel = noAuthNoPriv
        else:
            if privProtocol is None:
                maxSecurityLevel = authNoPriv
            else:
                maxSecurityLevel = authPriv
                kwargs["privProtocol"] = privProtocol
                kwargs["privSecret"] = privSecret or secret

            kwargs["authProtocol"] = authProtocol
            kwargs["authSecret"] = authSecret or secret

        if defaultSecurityLevel is None:
            defaultSecurityLevel = maxSecurityLevel
        elif defaultSecurityLevel > maxSecurityLevel:
            errmsg = "{} is required in order to support {}"
            param = "privProtocol" if maxSecurityLevel.auth else "authProtocol"
            raise ValueError(errmsg.format(param, defaultSecurityLevel))

        userName = userName.encode()

        with self.lock:
            try:
                space = self.namespaces[namespace]
            except KeyError:
                space = NameSpace(userName)
                self.namespaces[namespace] = space
            else:
                if userName in space:
                    errmsg = "User \"{}\" is already defined"

                    if namespace:
                        errmsg += " in namespace \"{}\"".format(namespace)

                    raise ValueError(errmsg.format(userName.decode()))

            if default:
                space.defaultUserName = userName

            space.addUser(userName, defaultSecurityLevel, kwargs)

    def connectTransport(self, transport):
        if transport.DOMAIN in self.transports:
            errmsg = "{} is already handled by a different transport object"
            raise ValueError(errmsg.format(transport.DOMAIN))
        elif transport.DOMAIN not in self.TRANSPORTS:
            raise ValueError(self.UNSUPPORTED.format(transport.DOMAIN))

        self.dispatcher.connectTransport(transport)
        self.transports.add(transport.DOMAIN)

    def v1Manager(self, locator):
        pass

    def v2cManager(self, locator, community, autowait=None):
        if autowait is None:
            autowait = self.autowaitDefault

        if locator.domain not in self.transports:
            transportClass = self.TRANSPORTS[locator.domain]
            self.dispatcher.connectTransport(transportClass())
            self.transports.add(locator.domain)

        if self.mpv2c is None:
            self.mpv2c = snmp.message.v2c.MessageProcessor(self.lockType)
            self.dispatcher.addMessageProcessor(self.mpv2c)

        return SNMPv2cManager(self, locator, community, autowait)

    def v3Manager(self, locator, securityModel=None, engineID=None,
            defaultUserName=None, namespace="", autowait=None):
        if securityModel is None:
            securityModel = self.defaultSecurityModel
        elif not isinstance(securityModel, SecurityModel):
            securityModel = SecurityModel(securityModel)

        try:
            space = self.namespaces[namespace]
        except KeyError as err:
            errmsg = f"No users defined in namespace \"{namespace}\""
            raise ValueError(errmsg) from err

        if defaultUserName is None:
            defaultUserName = space.defaultUserName
        else:
            defaultUserName = defaultUserName.encode()

        try:
            defaultUser = space.getUser(defaultUserName)
        except KeyError as err:
            errmsg = "No such user in namespace \"{}\": \"{}\""
            raise ValueError(errmsg.format(namespace, defaultUserName)) from err
        else:
            defaultSecurityLevel = defaultUser.defaultSecurityLevel

        if autowait is None:
            autowait = self.autowaitDefault

        if locator.domain not in self.transports:
            transportClass = self.TRANSPORTS[locator.domain]
            self.dispatcher.connectTransport(transportClass())
            self.transports.add(locator.domain)

        if self.mpv3 is None:
            self.mpv3 = snmp.message.v3.MessageProcessor(self.lockType)
            self.dispatcher.addMessageProcessor(self.mpv3)

        if securityModel == SecurityModel.USM:
            if self.usm is None:
                self.usm = SecurityModule(lockType=self.lockType)
                self.mpv3.secure(self.usm)

            return SNMPv3UsmManager(
                self,
                locator,
                namespace,
                defaultUserName,
                defaultSecurityLevel,
                engineID=engineID,
                autowait=autowait,
            )
        else:
            raise ValueError(self.UNSUPPORTED.format(str(securityModel)))

    def Manager(self, address, domain=None, version=None, **kwargs):
        if domain is None:
            domain = self.defaultDomain

        try:
            locator = self.TRANSPORTS[domain].Locator(address)
        except KeyError as err:
            raise ValueError(self.UNSUPPORTED.format(domain)) from err

        if version is None:
            version = self.defaultVersion
        elif not isinstance(version, MessageProcessingModel):
            version = MessageProcessingModel(version)

        if version == MessageProcessingModel.SNMPv3:
            return self.v3Manager(locator, **kwargs)
        elif version == MessageProcessingModel.SNMPv2c:
            return self.v2cManager(locator, **kwargs)
        elif version == MessageProcessingModel.SNMPv1:
            return self.v1Manager(locator, **kwargs)
        else:
            raise ValueError(self.UNSUPPORTED.format(str(version)))



__all__ = [
    "EncodeError", "ParseError",
    "CLASS_UNIVERSAL", "CLASS_APPLICATION",
    "CLASS_CONTEXT_SPECIFIC", "CLASS_PRIVATE",
    "STRUCTURE_PRIMITIVE", "STRUCTURE_CONSTRUCTED",
    "Identifier", "decode", "encode",
]

from collections import namedtuple
from .exception import *
from .utils import subbytes

class EncodeError(SNMPException):
    pass

class ParseError(IncomingMessageError):
    pass

CLASS_UNIVERSAL         = 0
CLASS_APPLICATION       = 1
CLASS_CONTEXT_SPECIFIC  = 2
CLASS_PRIVATE           = 3

STRUCTURE_PRIMITIVE     = 0
STRUCTURE_CONSTRUCTED   = 1

Identifier = namedtuple("Identifier", ("cls", "structure", "tag"))

def decode_identifier(data):
    try:
        byte = data.consume()
    except IndexError as err:
        raise ParseError("Missing identifier") from err

    cls         = (byte & 0xc0) >> 6
    structure   = (byte & 0x20) >> 5
    tag         = (byte & 0x1f)

    if tag == 0x1f:
        tag = 0
        byte = 0x80
        while byte & 0x80:
            try:
                byte = data.consume()
            except IndexError as err:
                raise ParseError("Incomplete identifier") from err

            tag <<= 7
            tag |= byte & 0x7f

    return Identifier(cls, structure, tag)

def encode_identifier(i):
    byte = (
        ((i.cls       << 6) & 0xc0) |
        ((i.structure << 5) & 0x20)
    )

    arr = bytearray()
    if i.tag < 0x1f:
        byte |= i.tag
    else:
        byte |= 0x1f

        tag = i.tag
        indicator = 0
        while tag:
            arr.append(indicator | (tag & 0x7f))
            indicator = 0x80
            tag >>= 7

    arr.append(byte)
    return bytes(reversed(arr))

def decode_length(data):
    try:
        length = data.consume()
    except IndexError as err:
        raise ParseError("Missing length") from err

    if length & 0x80:
        n = length & 0x7f

        length = 0
        for i in range(n):
            try:
                byte = data.consume()
            except IndexError as err:
                raise ParseError("Incomplete length") from err

            length <<= 8
            length |= byte

    return length

def encode_length(length):
    if length < 0x80:
        return bytes([length])

    arr = bytearray()
    while length:
        arr.append(length & 0xff)
        length >>= 8

    if len(arr) > 0x7f:
        raise EncodeError("Length too large for definite-length encoding")

    arr.append(0x80 | len(arr))
    return bytes(reversed(arr))

def decode(data, expected=None, leftovers=False, copy=True):
    data = subbytes(data)
    identifier = decode_identifier(data)

    result = []
    if expected is None:
        result.append(identifier)
    elif identifier != expected:
        raise ParseError("Identifier does not match expected type")

    length = decode_length(data)
    pruned = data.prune(length)

    if len(data) < length:
        raise ParseError("Incomplete value")

    if copy:
        result.append(data[:])
    else:
        result.append(data)

    if leftovers:
        result.append(pruned)
    elif pruned:
        raise ParseError("Trailing bytes")

    if len(result) == 1:
        return result[0]
    else:
        return tuple(result)

def encode(identifier, data):
    return encode_identifier(identifier) + encode_length(len(data)) + data



__all__ = [
    "InvalidEngineID", "InvalidUserName",
    "InvalidSecurityLevel", "SecurityModule",
]

from time import time
from snmp.ber import decode, encode
from snmp.exception import IncomingMessageError
from snmp.types import *
from snmp.security.levels import *
from snmp.utils import DummyLock, typename
from .. import SecurityModel, SecurityParameters

class UsmStatsError(IncomingMessageError):
    USM_STATS = OID.parse(".1.3.6.1.6.3.15.1.1")

    def __init__(self, msg):
        super().__init__(msg)
        self.oid = self.USM_STATS.extend(self.ERRNUM)

class UnsupportedSecLevel(UsmStatsError):
    ERRNUM = 1

class NotInTimeWindow(UsmStatsError):
    ERRNUM = 2

class UnknownUserName(UsmStatsError):
    ERRNUM = 3

class UnknownEngineID(UsmStatsError):
    ERRNUM = 4

class WrongDigest(UsmStatsError):
    ERRNUM = 5

class DecryptionError(UsmStatsError):
    ERRNUM = 6

class InvalidEngineID(ValueError):
    pass

class InvalidUserName(ValueError):
    pass

class InvalidSecurityLevel(ValueError):
    pass

class TimeEntry:
    def __init__(self, engineBoots, latestBootTime=None, authenticated=False):
        if latestBootTime is None:
            latestBootTime = time()

        self.authenticated = authenticated
        self.snmpEngineBoots = engineBoots
        self.latestBootTime = latestBootTime
        self.latestReceivedEngineTime = 0

    def snmpEngineTime(self, timestamp):
        return int(timestamp - self.latestBootTime)

class TimeKeeper:
    MAX_ENGINE_BOOTS = (1 << 31) - 1
    TIME_WINDOW_SIZE = 150

    def __init__(self, lockType):
        self.lock = lockType()
        self.table = {}

    def getEngineTime(self, engineID, timestamp=None):
        if timestamp is None:
            timestamp = time()

        with self.lock:
            try:
                entry = self.table[engineID]
            except KeyError:
                return 0, 0

            return entry.snmpEngineBoots, entry.snmpEngineTime(timestamp)

    def update(self, engineID, msgBoots=0, msgTime=0, timestamp=None):
        self.updateAndVerify(engineID, msgBoots, msgTime, False, timestamp)

    def updateAndVerify(self, engineID, msgBoots, msgTime,
                                auth=True, timestamp=None):
        if timestamp is None:
            timestamp = time()

        with self.lock:
            try:
                entry = self.table[engineID]
            except KeyError as err:
                entry = TimeEntry(msgBoots, timestamp - msgTime, auth)
                self.table[engineID] = entry

            withinTimeWindow = False
            if auth or not entry.authenticated:
                if msgBoots > entry.snmpEngineBoots:
                    entry.snmpEngineBoots = msgBoots
                    entry.latestBootTime = timestamp
                    entry.latestReceivedEngineTime = 0

                if msgBoots == entry.snmpEngineBoots:
                    if msgTime > entry.latestReceivedEngineTime:
                        entry.latestBootTime = timestamp - msgTime
                        entry.latestReceivedEngineTime = msgTime
                        withinTimeWindow = True
                    else:
                        snmpEngineTime = entry.snmpEngineTime(timestamp)
                        difference = snmpEngineTime - msgTime
                        if difference <= self.TIME_WINDOW_SIZE:
                            withinTimeWindow = True

                if entry.snmpEngineBoots == self.MAX_ENGINE_BOOTS:
                    withinTimeWindow = False

            if auth:
                entry.authenticated = True
                return withinTimeWindow

    # TODO: salvage this function
    def verifyTimeliness(self, engineID, msgBoots, msgTime, timestamp=None):
        if timestamp is None:
            timestamp = time()

        withinTimeWindow = False
        with self.lock:
            try:
                entry = self.table[engineID]
            except KeyError as err:
                raise InvalidEngineID(engineID) from err

            if msgBoots == entry.snmpEngineBoots:
                difference = entry.snmpEngineTime(timestamp) - msgTime
                if abs(difference) < self.TIME_WINDOW_SIZE:
                    withinTimeWindow = True

            if entry.snmpEngineBoots == self.MAX_ENGINE_BOOTS:
                withinTimeWindow = False

        return withinTimeWindow

class Credentials:
    def __init__(self, auth, priv):
        self.auth = auth
        self.priv = priv

class UserInfo:
    def __init__(self, engineID, name, auth=None, priv=None):
        self.engineID = engineID
        self.name = name
        self.credentials = Credentials(auth, priv)

class UserTable:
    def __init__(self, lockType):
        self.engines = {}
        self.lock = lockType()

    def addUser(self, userInfo):
        with self.lock:
            self.engines.setdefault(
                userInfo.engineID,
                dict()
            )[userInfo.name] = userInfo.credentials

    def getUser(self, engineID, userName):
        with self.lock:
            try:
                users = self.engines[engineID]
            except KeyError as err:
                raise InvalidEngineID(engineID) from err

            try:
                return users[userName]
            except KeyError as err:
                raise InvalidUserName(userName) from err

class SecurityModule:
    MODEL = SecurityModel.USM

    def __init__(self, lockType=DummyLock, engineID=None):
        self.engineID = engineID
        self.timekeeper = TimeKeeper(lockType)
        self.users = UserTable(lockType)

        if self.engineID is not None:
            self.timekeeper.update(self.engineID)

    def addUser(self, *args, **kwargs):
        self.users.addUser(self.makeCredentials(*args, **kwargs))

    @staticmethod
    def makeCredentials(engineID, userName,
            authProtocol=None, authKey=None,
            privProtocol=None, privKey=None):
        kwargs = dict()
        if authProtocol is not None:
            kwargs["auth"] = authProtocol(authKey)
            if privProtocol is not None:
                kwargs["priv"] = privProtocol(privKey)

        return UserInfo(engineID, userName, **kwargs)

    def prepareOutgoing(self, header, data, engineID,
                        securityName, securityLevel, userInfo=None):
        if securityLevel.auth:
            try:
                user = self.users.getUser(engineID, securityName)
            except (InvalidEngineID, InvalidUserName):
                if userInfo is None:
                    raise
                else:
                    user = userInfo.credentials

            if not user.auth:
                err = f"Authentication is disabled for user {securityName}"
                raise InvalidSecurityLevel(err)

            engineTimeParameters = self.timekeeper.getEngineTime(engineID)
            snmpEngineBoots, snmpEngineTime = engineTimeParameters
            msgAuthenticationParameters = user.auth.msgAuthenticationParameters
            msgPrivacyParameters = b''

            if securityLevel.priv:
                if not user.priv:
                    err = f"Privacy is disabled for user {securityName}"
                    raise InvalidSecurityLevel(err)

                msgPrivacyParameters, ciphertext = user.priv.encrypt(
                    data,
                    snmpEngineBoots,
                    snmpEngineTime,
                )

                data = OctetString(ciphertext).encode()

        else:
            if engineID == self.engineID:
                engineTimeParameters = self.timekeeper.getEngineTime(engineID)
                snmpEngineBoots, snmpEngineTime = engineTimeParameters
            else:
                snmpEngineBoots = 0
                snmpEngineTime = 0

            msgAuthenticationParameters = b''
            msgPrivacyParameters = b''

        encodedPrivacyParameters = OctetString(msgPrivacyParameters).encode()
        securityParameters = encode(
            SEQUENCE,
            b''.join((
                OctetString(engineID).encode(),
                Integer(snmpEngineBoots).encode(),
                Integer(snmpEngineTime).encode(),
                OctetString(securityName).encode(),
                OctetString(msgAuthenticationParameters).encode(),
                encodedPrivacyParameters,
            ))
        )

        msgSecurityParameters = OctetString(securityParameters).encode()
        body = b''.join((header, msgSecurityParameters, data))
        wholeMsg = encode(SEQUENCE, body)

        if securityLevel.auth:
            signature = user.auth.sign(wholeMsg)
            endIndex = len(wholeMsg) - len(data) - len(encodedPrivacyParameters)
            startIndex = endIndex - len(signature)
            wholeMsg = b''.join((
                wholeMsg[:startIndex],
                signature,
                wholeMsg[endIndex:]
            ))

        return wholeMsg

    def processIncoming(self, msg, securityLevel,
                        userInfo=None, timestamp=None):
        if timestamp is None:
            timestamp = time()

        msgSecurityParameters, msgData = \
            OctetString.decode(msg, leftovers=True, copy=False)
        ptr = decode(msgSecurityParameters.data, expected=SEQUENCE, copy=False)
        msgAuthoritativeEngineID, ptr = OctetString.decode(ptr, leftovers=True)
        msgAuthoritativeEngineBoots, ptr  = Integer.decode(ptr, leftovers=True)
        msgAuthoritativeEngineTime,  ptr  = Integer.decode(ptr, leftovers=True)
        msgUserName,              ptr = OctetString.decode(ptr, leftovers=True)

        msgAuthenticationParameters, ptr = \
            OctetString.decode(ptr, leftovers=True)
        msgAuthenticationParametersIndex = \
            ptr.start - len(msgAuthenticationParameters.data)
        msgPrivacyParameters = OctetString.decode(ptr)

        engineID = msgAuthoritativeEngineID.data
        userName = msgUserName.data
        securityParameters = SecurityParameters(engineID, userName)

        if not securityLevel.auth:
            self.timekeeper.update(
                engineID,
                msgAuthoritativeEngineBoots.value,
                msgAuthoritativeEngineTime.value,
                timestamp=timestamp
            )

            return securityParameters, msgData[:]

        addUser = False

        try:
            user = self.users.getUser(engineID, userName)
        except InvalidEngineID as err:
            if userInfo is None or engineID != userInfo.engineID:
                raise UnknownEngineID(engineID) from err
            elif userName != userInfo.name:
                raise UnknownUserName(userName) from err
            else:
                user = userInfo.credentials
                addUser = True
        except InvalidUserName as err:
            if userInfo is None or userName != userInfo.name:
                raise UnknownUserName(userName) from err
            else:
                user = userInfo.credentials
                addUser = True

        if user.auth is None:
            err = f"Authentication is disabled for user {userName}"
            raise UnsupportedSecLevel(err)
        elif securityLevel.priv and user.priv is None:
            err = f"Data privacy is disabled for user {userName}"
            raise UnsupportedSecLevel(err)

        padding = user.auth.msgAuthenticationParameters
        if len(msgAuthenticationParameters.data) != len(padding):
            raise WrongDigest("Invalid signature length")

        wholeMsg = b''.join((
            msg.data[:msgAuthenticationParametersIndex],
            padding,
            msg.data[msgAuthenticationParametersIndex + len(padding):]
        ))

        if user.auth.sign(wholeMsg) != msgAuthenticationParameters.data:
            raise WrongDigest("Invalid signature")

        try:
            if not self.timekeeper.updateAndVerify(
                engineID,
                msgAuthoritativeEngineBoots.value,
                msgAuthoritativeEngineTime.value,
                timestamp=timestamp,
            ):
                raise NotInTimeWindow(
                    engineID,
                    msgAuthoritativeEngineBoots.value,
                    msgAuthoritativeEngineTime.value,
                )
        except InvalidEngineID as err:
            raise UnknownEngineID(engineID) from err

        if securityLevel.priv:
            try:
                payload = user.priv.decrypt(
                    OctetString.decode(msgData).data,
                    msgAuthoritativeEngineBoots.value,
                    msgAuthoritativeEngineTime.value,
                    msgPrivacyParameters.data
                )
            except ValueError as err:
                raise DecryptionError(str(err)) from err
        else:
            payload = msgData[:]

        if addUser:
            self.users.addUser(userInfo)

        return securityParameters, payload



__all__ = ["TransportDomain"]

from collections import namedtuple
import enum
import os
from snmp.exception import IncompleteChildClass
from snmp.utils import typename

class Transport:
    class Listener:
        def hear(self, transport, address, data):
            pass

    @classmethod
    def Locator(cls, address):
        return TransportLocator(cls.DOMAIN, cls.normalizeAddress(address))

    @classmethod
    def normalizeAddress(cls, address):
        return address

    def close(self):
        pass

    def listen(self, listener):
        pass

    def send(self, address, data):
        errmsg = "{} does not implement send()".format(typename(self))
        raise IncompleteChildClass(errmsg)

    def stop(self):
        pass

TransportDomain = enum.Enum("TransportDomain", ("UDP",))
TransportLocator = namedtuple("TransportLocator", ("domain", "address"))

supported = ("nt", "posix")
if os.name in supported:
    package = os.name
else:
    from platform import platform
    raise ImportError("Unsupported platform: \"{}\"".format(platform()))

package = "{}.{}".format(__name__, package)



class SNMPException(Exception):
    pass

class SNMPLibraryBug(AssertionError):
    pass

class IncompleteChildClass(SNMPLibraryBug):
    pass

class UnsupportedFeature(SNMPLibraryBug):
    pass

class IncomingMessageError(SNMPException):
    pass



__all__ = ["SecurityLevel", "SecurityModel", "SecurityParameters"]

import enum
from snmp.utils import typename

class SecurityLevel:
    def __init__(self, auth=False, priv=False):
        self._auth = False
        self._priv = False

        self.auth = auth
        self.priv = priv

    def __repr__(self):
        return "{}(auth={}, priv={})".format(
            typename(self),
            self.auth,
            self.priv
        )

    def __str__(self):
        return "{}{}".format(
            "auth" if self.auth else "noAuth",
            "Priv" if self.priv else "NoPriv"
        )

    @property
    def auth(self):
        return self._auth

    @property
    def priv(self):
        return self._priv

    @auth.setter
    def auth(self, value):
        if not value and self.priv:
            msg = "Cannot disable authentication while privacy is enabled"
            raise ValueError(msg)

        self._auth = bool(value)

    @priv.setter
    def priv(self, value):
        if value and not self.auth:
            msg = "Cannot enable privacy while authentication is disabled"
            raise ValueError(msg)

        self._priv = bool(value)

    def __eq__(a, b):
        return a.auth == b.auth and a.priv == b.priv

    def __lt__(a, b):
        if a.auth:
            return b.priv and not a.priv
        else:
            return b.auth

    def __ge__(a, b):
        return not a < b

class SecurityModel(enum.IntEnum):
    USM = 3

class SecurityParameters:
    def __init__(self, engineID, userName):
        self.securityEngineID = engineID
        self.securityName = userName

    def __repr__(self):
        return f"{typename(self)}({self.securityEngineID}, {self.securityName})"



__all__ = ["DesCbc"]

import os
from snmp.openssl.des import ffi, lib
from . import DecryptionError

class DesCbc:
    BYTEORDER = "big"
    BLOCKLEN = ffi.sizeof("DES_cblock")
    KEYLEN = BLOCKLEN * 2
    SALTLEN = BLOCKLEN // 2
    SALTWRAP = 1 << (8 * SALTLEN)

    def __init__(self, key):
        if len(key) < self.KEYLEN:
            errmsg = "key must be at least {} bytes long".format(self.KEYLEN)
            raise ValueError(errmsg)

        key, self.preIV = (
            ffi.new("DES_cblock*", key[:self.BLOCKLEN]),
            ffi.new("DES_cblock*", key[self.BLOCKLEN:self.KEYLEN])
        )

        self.key = ffi.new("DES_key_schedule*")
        lib.DES_set_odd_parity(key)
        lib.DES_set_key_unchecked(key, self.key)

        self.salt = int.from_bytes(os.urandom(self.SALTLEN), self.BYTEORDER)

    def compute(self, data, salt, mode):
        iv = ffi.new("DES_cblock*")
        ffi.memmove(iv, self.preIV, self.BLOCKLEN)

        for i in range(self.BLOCKLEN):
            iv[0][i] ^= salt[i]

        output = ffi.new("unsigned char[]", len(data))
        lib.DES_cbc_encrypt(data, output, len(data), self.key, iv, mode)
        return bytes(output)

    def pad(self, data):
        n = self.BLOCKLEN - (len(data) % self.BLOCKLEN)
        return (data + (b'\0' * n)) if n else data

    def decrypt(self, data, engineBoots, engineTime, salt):
        if len(data) % self.BLOCKLEN:
            errmsg = "DES ciphertext must be a multiple of {} in length"
            raise DecryptionError(errmsg.format(self.BLOCKLEN))

        return self.compute(data, salt, lib.DES_DECRYPT)

    def encrypt(self, data, engineBoots, engineTime):
        self.salt = (self.salt + 1) % self.SALTWRAP
        salt = b''.join((
            engineBoots.to_bytes(self.BLOCKLEN - self.SALTLEN, self.BYTEORDER),
            self.salt.to_bytes(self.SALTLEN, self.BYTEORDER),
        ))

        return salt, self.compute(self.pad(data), salt, lib.DES_ENCRYPT)



import weakref

from snmp.ber import *
from snmp.ber import ParseError, decode_identifier
from snmp.exception import *
from snmp.message import *
from snmp.pdu.v2 import *
from snmp.types import *
from snmp.utils import *
from . import *

class LateResponse(IncomingMessageError):
    pass

class ResponseMismatch(IncomingMessageError):
    @classmethod
    def byField(cls, field):
        return cls(f"{field} does not match request")

class CacheEntry:
    def __init__(self, handle, community):
        self.community = community
        self.handle = handle

class SNMPv2cMessage(Sequence):
    VERSION = MessageProcessingModel.SNMPv2c

    def __init__(self, community, pdu):
        self.community = community
        self.pdu = pdu

    def __repr__(self):
        return f"{typename(self)}({self.community}, {repr(self.pdu)})"

    def __str__(self, depth=0, tab="    "):
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Community: {self.community.pdu}",
            f"{self.pdu.__str__(depth+1, tab)}",
        ))

    @property
    def objects(self):
        yield Integer(self.VERSION)
        yield OctetString(self.community)
        yield self.pdu

class MessageProcessor:
    VERSION = SNMPv2cMessage.VERSION

    def __init__(self, lockType=DummyLock):
        self.cacheLock = lockType()
        self.generator = NumberGenerator(32)
        self.outstanding = {}

    def cache(self, entry, credentials=None):
        retry = 0
        while retry < 10:
            with self.cacheLock:
                requestID = next(self.generator)
                if requestID not in self.outstanding:
                    self.outstanding[requestID] = entry
                    return requestID

            retry += 1

        raise Exception("Failed to allocate request ID")

    def retrieve(self, requestID):
        with self.cacheLock:
            return self.outstanding[requestID]

    def uncache(self, requestID):
        with self.cacheLock:
            try:
                del self.outstanding[requestID]
            except KeyError:
                pass

    def prepareDataElements(self, msg):
        community, msg = OctetString.decode(msg, leftovers=True)
        identifier = decode_identifier(subbytes(msg))

        try:
            pduType = pduTypes[identifier]
        except KeyError as err:
            raise ParseError(f"Invalid PDU type: {identifier}") from err

        pdu = pduType.decode(msg)
        if isinstance(pdu, ResponsePDU):
            try:
                entry = self.retrieve(pdu.requestID)
            except KeyError as err:
                errmsg = f"Unknown requestID: {pdu.requestID}"
                raise ResponseMismatch(errmsg) from err

            handle = entry.handle()
            if handle is None:
                raise LateResponse("Handle has already been released")

            if entry.community != community.data:
                raise ResponseMismatch.byField("Community")
        else:
            raise UnsupportedFeature("Received a non-response PDU type")

        return SNMPv2cMessage(community, pdu), handle

    def prepareOutgoingMessage(self, pdu, handle, community):
        pdu.requestID = self.cache(CacheEntry(weakref.ref(handle), community))
        handle.addCallback(self.uncache, pdu.requestID)

        return SNMPv2cMessage(community, pdu).encode()



__all__ = ["Repeater", "Timeout"]

import math
import time
import weakref

from snmp.exception import *

class Timeout(SNMPException):
    pass

class Repeater:
    def __init__(self, request, period=1.0, timeout=10.0):
        now = time.time()

        self.expiration = now + timeout
        self.nextRefresh = now + period
        self.period = period
        self.request = weakref.ref(request)

    def __lt__(a, b):
        return a.target < b.target

    def refresh(self):
        request = self.request()
        if request is None or request.fulfilled:
            return None

        now = time.time()
        if self.expiration <= now:
            request.expired = True
            return 0.0

        delta = self.target - now

        if delta < 0:
            self.nextRefresh += math.ceil(-delta / self.period) * self.period
            #print(f"Refreshing {id(self)} @ {now} ; nextRefresh = {self.nextRefresh} ; expiration = {self.expiration}")
            request.send()

        return delta

    @property
    def target(self):
        return min(self.expiration, self.nextRefresh)



__all__ = ["Dispatcher"]

# TODO: this module tightly depends on the threading module
#       in order to support other concurrency models (e.g. async),
#       this dependency will need to be broken
import threading
from snmp.ber import ParseError, decode
from snmp.exception import *
from snmp.message import MessageProcessingModel
from snmp.security.levels import noAuthNoPriv
from snmp.transport import Transport, TransportDomain
from snmp.types import SEQUENCE, Integer
from snmp.utils import DummyLock, typename

class Dispatcher(Transport.Listener):
    class Handle:
        def addCallback(self, func, *args):
            errmsg = "{} does not support callbacks".format(typename(self))
            raise IncompleteChildClass(errmsg)

        def push(self, response):
            errmsg = "{} does not implement push()".format(typename(self))
            raise IncompleteChildClass(errmsg)

    def __init__(self, lockType=DummyLock):
        self.lock = lockType()
        self.msgProcessors = {}
        self.threads = {}
        self.transports = {}

    def addMessageProcessor(self, mp):
        with self.lock:
            self.msgProcessors[mp.VERSION] = mp

    def connectTransport(self, transport):
        domain = transport.DOMAIN

        with self.lock:
            if domain in self.threads:
                if self.transports.get(domain) is transport:
                    return

                errmsg = "This {} instance is already connected to {}"
                raise ValueError(errmsg.format(typename(self), str(domain)))

            thread = threading.Thread(
                target=transport.listen,
                args=(self,))

            thread.start()

            self.transports[domain] = transport
            self.threads[domain] = thread

    def hear(self, transport, address, data):
        try:
            try:
                message = decode(data, expected=SEQUENCE, copy=False)
                msgVersion, message = Integer.decode(message, leftovers=True)
            except ParseError:
                return

            with self.lock:
                try:
                    mp = self.msgProcessors[msgVersion.value]
                except KeyError:
                    return

            try:
                message, handle = mp.prepareDataElements(message)
            except IncomingMessageError:
                return

            handle.push(message)
        except AssertionError:
            pass
        except Exception:
            pass

    def sendPdu(self, domain, address, mpm, pdu, handle, *args, **kwargs):
        with self.lock:
            try:
                transport = self.transports[domain]
            except KeyError as err:
                domain = str(TransportDomain(domain))
                raise ValueError("{} is not enabled".format(domain)) from err

            try:
                mp = self.msgProcessors[mpm]
            except KeyError as err:
                mpm = str(MessageProcessingModel(mpm))
                raise ValueError("{} is not enabled".format(mpm)) from err

        msg = mp.prepareOutgoingMessage(pdu, handle, *args, **kwargs)
        transport.send(address, msg)
        return handle

    def shutdown(self):
        with self.lock:
            for transport in self.transports.values():
                transport.stop()

            for thread in self.threads.values():
                thread.join()

            for transport in self.transports.values():
                transport.close()

            self.transports.clear()
            self.threads.clear()



import heapq
import threading

from snmp.message import *
from snmp.pdu.v2 import *
from . import *

class Request:
    def __init__(self, pdu, manager, community):
        self.community = community
        self.manager = manager
        self.pdu = pdu

        self.callback = None
        self.messages = set()

        self.event = threading.Event()
        self.expired = False
        self.fulfilled = False
        self.response = None

    def __del__(self):
        if self.callback is not None:
            self.close()

    def close(self):
        for message in self.messages:
            self.callback(message)

        self.callback = None
        self.messages.clear()

    def addCallback(self, callback, msgID):
        if self.callback is None:
            self.callback = callback

        assert self.callback == callback
        self.messages.add(msgID)

    def push(self, response, depth=0):
        self.response = response
        self.event.set()

    def send(self):
        pdu = self.pdu
        community = self.community
        return self.manager.sendPdu(pdu, self, community)

    def wait(self):
        pdu = None
        while not self.expired:
            timeout = self.manager.refresh()
            if self.event.wait(timeout=timeout):
                pdu = self.response.pdu
                self.fulfilled = True
                break

        self.close()
        if pdu is None:
            raise Timeout()
        else:
            return pdu

class SNMPv2cManager:
    def __init__(self, engine, locator, community, autowait=True):
        self.autowait = autowait
        self.locator = locator

        self.localEngine = engine
        self.defaultCommunity = community

        self.lock = threading.Lock()
        self.requests = []

    def refresh(self):
        while True:
            with self.lock:
                entry = self.requests[0]
                result = entry.refresh()

                while result is None:
                    heapq.heappop(self.requests)
                    entry = self.requests[0]
                    result = entry.refresh()

                if result < 0:
                    heapq.heapreplace(self.requests, entry)
                else:
                    return result

    def sendPdu(self, pdu, handle, community):
        return self.localEngine.dispatcher.sendPdu(
            self.locator.domain,
            self.locator.address,
            MessageProcessingModel.SNMPv2c,
            pdu,
            handle,
            community
        )

    def sendRequest(self, pdu, community=None, wait=None, **kwargs):
        if community is None:
            community = self.defaultCommunity

        if wait is None:
            wait = self.autowait

        request = Request(pdu, self, community, **kwargs)

        with self.lock:
            heapq.heappush(self.requests, Repeater(request))
            request.send()

        if wait:
            return request.wait()
        else:
            return request

    def get(self, *oids, **kwargs):
        pdu = GetRequestPDU(*oids)
        return self.sendRequest(pdu, **kwargs)

    def getBulk(self, *oids, nonRepeaters=0, maxRepetitions=0, **kwargs):
        pdu = GetBulkRequestPDU(
            *oids,
            nonRepeaters=nonRepeaters,
            maxRepetitions=maxRepetitions,
        )

        return self.sendRequest(pdu, **kwargs)

    def getNext(self, *oids, **kwargs):
        pdu = GetNextRequestPDU(*oids)
        return self.sendRequest(pdu, **kwargs)

    def set(self, *varbinds, **kwargs):
        varbinds = (VarBind(*varbind) for varbind in varbinds)
        pdu = SetRequestPDU(*varbinds)
        return self.sendRequest(pdu, **kwargs)



__all__ = ["DummyLock", "NumberGenerator", "subbytes", "typename"]

from random import randint

class DummyLock:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def acquire(self, *args, **kwargs):
        return True

    def release(self):
        pass

class NumberGenerator:
    def __init__(self, nbits, signed=True):
        half = 1 << (nbits-1)

        self.previous = 0
        self.range = half << 1
        self.step = 2 * randint(1, half) - 1
        self.wrap = self.range - 1

        if signed:
            self.wrap -= half

    def __iter__(self):
        return self

    def __next__(self):
        self.previous += self.step

        if self.previous > self.wrap:
            self.previous -= self.range

        return self.previous

class subbytes:
    def __init__(self, data, start=None, stop=None):
        if isinstance(data, subbytes):
            self.data = data.data
        else:
            self.data = data
            self.start = 0
            self.stop = len(data)
            data = self

        new_start = data.start if start is None else data.translate(start, True)
        new_stop  = data.stop  if stop  is None else data.translate(stop,  True)

        self.start = new_start
        self.stop  = new_stop

    def __bool__(self):
        return self.stop > self.start

    def __eq__(a, b):
        if len(a) != len(b):
            return False

        for left, right in zip(a, b):
            if left != right:
                return False

        return True

    def __iter__(self):
        for index in range(self.start, self.stop):
            yield self.data[index]

    def __len__(self):
        return self.stop - self.start

    def __repr__(self):
        args = [repr(self.data)]

        if self.start:
            args.append(f"start={self.start}")

        if self.stop < len(self.data):
            args.append(f"stop={self.stop}")

        return f"{typename(self)}({', '.join(args)})"

    def translate(self, index, clamp=False):
        if index < 0:
            index += self.stop
            if index < self.start:
                if clamp:
                    return self.start
                else:
                    return len(self.data)
            else:
                return index
        else:
            index += self.start
            if index >= self.stop:
                if clamp:
                    return self.stop
                else:
                    return len(self.data)
            else:
                return index

    def __getitem__(self, key):
        if isinstance(key, int):
            key = self.translate(key)
        elif isinstance(key, slice):
            start, stop = key.start, key.stop

            key = slice(
                self.start if start is None else self.translate(start, True),
                self.stop  if stop  is None else self.translate(stop, True),
                key.step
            )

        return self.data[key]

    def consume(self):
        byte = self.dereference()
        self.start += 1
        return byte

    def dereference(self):
        return self[0]

    def prune(self, length):
        removed = subbytes(self, length)
        self.stop = removed.start
        return removed

def typename(cls, qualified=False):
    if not isinstance(cls, type):
        cls = type(cls)

    if qualified:
        return ".".join((cls.__module__, cls.__qualname__))
    else:
        return cls.__name__



import weakref

from snmp.ber import ParseError, decode_identifier
from snmp.exception import *
from snmp.pdu.v2 import *
from snmp.security import *
from snmp.security.levels import noAuthNoPriv
from snmp.types import *
from snmp.utils import *
from . import MessageProcessingModel

class InvalidMessage(IncomingMessageError):
    pass

class LateResponse(IncomingMessageError):
    pass

class ResponseMismatch(IncomingMessageError):
    @classmethod
    def byField(cls, field):
        return cls(f"{field} does not match request")

class UnknownSecurityModel(IncomingMessageError):
    pass

class MessageFlags(OctetString):
    MIN_SIZE = 1

    AUTH_FLAG       = (1 << 0)
    PRIV_FLAG       = (1 << 1)
    REPORTABLE_FLAG = (1 << 2)
    ALL_FLAGS       = AUTH_FLAG | PRIV_FLAG | REPORTABLE_FLAG

    def __init__(self, byte=0):
        self.byte = byte & self.ALL_FLAGS

    def __repr__(self):
        return f"{typename(self)}({self.byte})"

    def __str__(self):
        flags = []
        if self.authFlag:
            flags.append("AUTH")

        if self.privFlag:
            flags.append("PRIV")

        if self.reportableFlag:
            flags.append("REPORTABLE")

        return f"<{','.join(flags)}>"

    @classmethod
    def parse(cls, data=b''):
        return cls(byte=data[0])

    @property
    def data(self):
        return bytes((self.byte,))

    @property
    def authFlag(self):
        return bool(self.byte & self.AUTH_FLAG)

    @property
    def privFlag(self):
        return bool(self.byte & self.PRIV_FLAG)

    @property
    def reportableFlag(self):
        return bool(self.byte & self.REPORTABLE_FLAG)

    @authFlag.setter
    def authFlag(self, value):
        if value:
            self.byte |= self.AUTH_FLAG
        else:
            self.byte &= ~self.AUTH_FLAG

    @privFlag.setter
    def privFlag(self, value):
        if value:
            self.byte |= self.PRIV_FLAG
        else:
            self.byte &= ~self.PRIV_FLAG

    @reportableFlag.setter
    def reportableFlag(self, value):
        if value:
            self.byte |= self.REPORTABLE_FLAG
        else:
            self.byte &= ~self.REPORTABLE_FLAG

class HeaderData(Sequence):
    def __init__(self, msgID, maxSize, flags, securityModel):
        self.id = msgID
        self.maxSize = maxSize
        self.flags = flags
        self.securityModel = securityModel

    def __repr__(self):
        args = (
            str(self.id),
            str(self.maxSize),
            repr(self.flags),
            str(SecurityModel(self.securityModel)),
        )

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self, depth=0, tab="    "):
        indent = tab * depth
        subindent = indent + tab
        securityModel = SecurityModel(self.securityModel)
        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Message ID: {self.id}",
            f"{subindent}Sender Message Size Limit: {self.maxSize}",
            f"{subindent}Flags: {self.flags}",
            f"{subindent}Security Model: {securityModel.name}"
        ))

    @property
    def objects(self):
        yield Integer(self.id)
        yield Integer(self.maxSize)
        yield self.flags
        yield Integer(self.securityModel)

    @classmethod
    def deserialize(cls, data):
        msgID,      data = Integer      .decode(data, leftovers=True)
        msgMaxSize, data = Integer      .decode(data, leftovers=True)
        msgFlags,   data = MessageFlags .decode(data, leftovers=True)
        msgSecurityModel = Integer      .decode(data)

        if msgID.value < 0:
            raise ParseError("msgID may not be less than 0")
        elif msgMaxSize.value < 484:
            raise ParseError("msgMaxSize may not be less than 484")
        elif msgSecurityModel.value < 1:
            raise ParseError("msgSecurityModel may not be less than 1")

        return cls(
            msgID.value,
            msgMaxSize.value,
            msgFlags,
            msgSecurityModel.value
        )

class ScopedPDU(Sequence):
    def __init__(self, pdu, contextEngineID, contextName=b''):
        self.contextEngineID = contextEngineID
        self.contextName = contextName
        self.pdu = pdu

    def __repr__(self):
        args = (
            repr(self.pdu),
            repr(self.contextEngineID),
            f"contextName={repr(self.contextName)}"
        )

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self, depth=0, tab="    "):
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Context Engine ID: {self.contextEngineID}",
            f"{subindent}Context Name: {self.contextName}",
            f"{self.pdu.__str__(depth=depth+1, tab=tab)}"
        ))

    @property
    def objects(self):
        yield OctetString(self.contextEngineID)
        yield OctetString(self.contextName)
        yield self.pdu

    @classmethod
    def deserialize(cls, data, types={}):
        contextEngineID, data = OctetString.decode(data, leftovers=True)
        contextName,     data = OctetString.decode(data, leftovers=True)

        identifier = decode_identifier(subbytes(data))

        try:
            pduType = types[identifier]
        except KeyError as err:
            raise ParseError(f"Invalid PDU type: {identifier}") from err

        return cls(
            pduType.decode(data),
            contextEngineID=contextEngineID.data,
            contextName=contextName.data,
        )

class CacheEntry:
    def __init__(self, engineID, contextName, handle,
            securityName, securityModel, securityLevel):
        self.context = contextName
        self.engineID = engineID
        self.handle = handle
        self.securityName = securityName
        self.securityModel = securityModel
        self.securityLevel = securityLevel

class SNMPv3Message:
    def __init__(self, msgID, securityLevel, securityParameters, data):
        self.id = msgID
        self.securityLevel = securityLevel
        self.securityEngineID = securityParameters.securityEngineID
        self.securityName = securityParameters.securityName
        self.data = data

    def __repr__(self):
        args = (repr(member) for member in (
            self.id,
            self.securityLevel,
            SecurityParameters(self.securityEngineID, self.securityName),
            self.data,
        ))

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self, depth=0, tab="    "):
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Message ID: {self.id}",
            f"{subindent}Security Engine ID: {self.securityEngineID}",
            f"{subindent}Security Level: {self.securityLevel}",
            f"{subindent}Security Name: {self.securityName}",
            f"{self.data.__str__(depth+1, tab)}",
        ))

class MessageProcessor:
    VERSION = MessageProcessingModel.SNMPv3

    def __init__(self, lockType=DummyLock):
        self.cacheLock = lockType()
        self.credentials = {}
        self.generator = NumberGenerator(31, signed=False)
        self.outstanding = {}

        self.securityLock = lockType()
        self.defaultSecurityModel = None
        self.securityModules = {}

    def cache(self, entry, credentials=None):
        retry = 0
        while retry < 10:
            with self.cacheLock:
                msgID = next(self.generator)
                if msgID not in self.outstanding:
                    if credentials is not None:
                        self.credentials[msgID] = credentials

                    self.outstanding[msgID] = entry
                    return msgID

            retry += 1

        raise Exception("Failed to allocate message ID")

    def retrieve(self, msgID):
        with self.cacheLock:
            return self.outstanding[msgID]

    def uncache(self, msgID):
        with self.cacheLock:
            try:
                del self.outstanding[msgID]
            except KeyError:
                pass

    def secure(self, module, default=False):
        with self.securityLock:
            self.securityModules[module.MODEL] = module

            if default or self.defaultSecurityModel is None:
                self.defaultSecurityModel = module.MODEL

    def prepareDataElements(self, msg):
        msgGlobalData, msg = HeaderData.decode(msg, leftovers=True)

        with self.securityLock:
            try:
                securityModule = self.securityModules[
                    msgGlobalData.securityModel
                ]
            except KeyError as e:
                raise UnknownSecurityModel(msgGlobalData.securityModel) from e

        try:
            securityLevel = SecurityLevel(
                auth=msgGlobalData.flags.authFlag,
                priv=msgGlobalData.flags.privFlag)
        except ValueError as err:
            raise InvalidMessage(f"Invalid msgFlags: {err}") from err

        with self.cacheLock:
            credentials = self.credentials.get(msgGlobalData.id)

        security, data = \
            securityModule.processIncoming(msg, securityLevel, credentials)

        scopedPDU, _ = ScopedPDU.decode(data, types=pduTypes, leftovers=True)

        if isinstance(scopedPDU.pdu, Response):
            try:
                entry = self.retrieve(msgGlobalData.id)
            except KeyError as err:
                errmsg = f"Unknown msgID: {msgGlobalData.id}"
                raise ResponseMismatch(errmsg) from err

            handle = entry.handle()
            if handle is None:
                raise LateResponse("Handle has already been released")

            report = isinstance(scopedPDU.pdu, Internal)
            if not report and entry.securityLevel < securityLevel:
                raise ResponseMismatch.byField("Security Level")

            if not report and entry.engineID != security.securityEngineID:
                raise ResponseMismatch.byField("Security Engine ID")

            if entry.securityName != security.securityName:
                raise ResponseMismatch.byField("Security Name")

            if not report and entry.engineID != scopedPDU.contextEngineID:
                raise ResponseMismatch.byField("Context Engine ID")

            if entry.context != scopedPDU.contextName:
                raise ResponseMismatch.byField("Context Name")
        else:
            raise UnsupportedFeature("Received a non-response PDU type")

        message = \
            SNMPv3Message(msgGlobalData.id, securityLevel, security, scopedPDU)
        return message, handle

    def prepareOutgoingMessage(self, pdu, handle, engineID, securityName,
            securityLevel=noAuthNoPriv, securityModel=None, credentials=None,
            contextName=b''):

        with self.securityLock:
            if securityModel is None:
                securityModel = self.defaultSecurityModel

            try:
                securityModule = self.securityModules[securityModel]
            except KeyError as err:
                errmsg = f"Security Model {securityModel} has not been enabled"
                raise ValueError(errmsg) from err

        entry = CacheEntry(
            engineID,
            contextName,
            weakref.ref(handle),
            securityName,
            securityModel,
            securityLevel)

        msgID = self.cache(entry, credentials=credentials)
        handle.addCallback(self.uncache, msgID)

        flags = MessageFlags()
        flags.authFlag = securityLevel.auth
        flags.privFlag = securityLevel.priv
        flags.reportableFlag = isinstance(pdu, Confirmed)

        msgGlobalData = HeaderData(msgID, 1472, flags, securityModel)
        header = Integer(self.VERSION).encode() + msgGlobalData.encode()
        scopedPDU = ScopedPDU(pdu, engineID, contextName=contextName)

        return securityModule.prepareOutgoing(
            header,
            scopedPDU.encode(),
            engineID,
            securityName,
            securityLevel,
            credentials,
        )



__all__ = ["UdpTransport"]

import os
import select
import socket
from ..udp import UdpTransportBase

RECV_SIZE = 65507

class UdpTransport(UdpTransportBase):
    def __init__(self, host="", port=0):
        self.pipe = os.pipe()
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setblocking(False)
        self.socket.bind((host, port))

    def close(self):
        os.close(self.pipe[0])
        os.close(self.pipe[1])
        self.socket.close()

    def listen(self, listener):
        abort = self.pipe[0]
        sock = self.socket.fileno()

        poller = select.poll()
        poller.register(sock, select.POLLIN)
        poller.register(abort, select.POLLIN)

        done = False
        while not done:
            ready = poller.poll()
            for fd, _ in ready:
                if fd == sock:
                    data, addr = self.socket.recvfrom(RECV_SIZE)
                    listener.hear(self, addr, data)
                elif fd == abort:
                    os.read(fd, 1)
                    done = True

    def send(self, addr, packet):
        self.socket.sendto(packet, addr)

    def stop(self):
        os.write(self.pipe[1], b'\0')



__all__ = [
    "NoSuchObject", "NoSuchInstance", "EndOfMibView",
    "VarBind", "VarBindList", "PDU", "pduTypes",
    "GetRequestPDU", "GetNextRequestPDU", "ResponsePDU", "SetRequestPDU",
    "GetBulkRequestPDU", "InformRequestPDU", "TrapPDU", "ReportPDU",
    "Read", "Write", "Response", "Internal", "Notification", "Confirmed",
]

import enum
from snmp.ber import *
from snmp.smi.v2 import *
from snmp.types import *
from snmp.utils import subbytes, typename

class NoSuchObject(Null):
    TYPE = Identifier(CLASS_CONTEXT_SPECIFIC, STRUCTURE_PRIMITIVE, 0)

class NoSuchInstance(Null):
    TYPE = Identifier(CLASS_CONTEXT_SPECIFIC, STRUCTURE_PRIMITIVE, 1)

class EndOfMibView(Null):
    TYPE = Identifier(CLASS_CONTEXT_SPECIFIC, STRUCTURE_PRIMITIVE, 2)

class VarBind(Sequence):
    TYPES = {
        cls.TYPE: cls for cls in (
            Integer,
            OctetString,
            Null,
            OID,
            IpAddress,
            Counter32,
            Unsigned32,
            TimeTicks,
            Opaque,
            Counter64,
            NoSuchObject,
            NoSuchInstance,
            EndOfMibView,
        )
    }

    def __init__(self, name, value=None):
        if not isinstance(name, OID):
            name = OID.parse(name)

        if value is None:
            value = Null()

        self.name = name
        self.value = value

    def __iter__(self):
        return self.objects

    def __len__(self):
        return 2

    def __repr__(self):
        args = ", ".join((repr(self.name), repr(self.value)))
        return "{}({})".format(typename(self), args)

    def __str__(self):
        return "{}: {}".format(self.name, self.value)

    @property
    def objects(self):
        yield self.name
        yield self.value

    @classmethod
    def deserialize(cls, data):
        name, data = OID.decode(data, leftovers=True)
        identifier, data = decode(subbytes(data))

        try:
            valueType = cls.TYPES[identifier]
        except KeyError as err:
            msg = "Invalid variable value type: {}"
            raise ParseError(msg.format(identifier)) from err

        return cls(name, valueType.deserialize(data))

class VarBindList(Sequence):
    def __init__(self, *args):
        self.variables = [None] * len(args)
        for i, var in enumerate(args):
            if not isinstance(var, VarBind):
                var = VarBind(var)
            self.variables[i] = var

    def __bool__(self):
        return bool(self.variables)

    def __getitem__(self, key):
        return self.variables[key]

    def __iter__(self):
        return iter(self.variables)

    def __len__(self):
        return len(self.variables)

    def __repr__(self):
        args = ", ".join(repr(var) for var in self.variables)
        return "{}({})".format(typename(self), args)

    def __str__(self, indent=""):
        return "\n".join("{}{}".format(indent, var) for var in self.variables)

    @property
    def objects(self):
        return self.variables

    @classmethod
    def deserialize(cls, data):
        objects = []

        while data:
            var, data = VarBind.decode(data, leftovers=True)
            objects.append(var)

        return cls(*objects)

class PDU(Constructed):
    class ErrorStatus(enum.IntEnum):
        noError             = 0
        tooBig              = enum.auto()
        noSuchName          = enum.auto()
        badValue            = enum.auto()
        readOnly            = enum.auto()
        genErr              = enum.auto()
        noAccess            = enum.auto()
        wrongType           = enum.auto()
        wrongLength         = enum.auto()
        wrongEncoding       = enum.auto()
        wrongValue          = enum.auto()
        noCreation          = enum.auto()
        inconsistentValue   = enum.auto()
        resourceUnavailable = enum.auto()
        commitFailed        = enum.auto()
        undoFailed          = enum.auto()
        authorizationError  = enum.auto()
        notWritable         = enum.auto()
        inconsistentName    = enum.auto()

    def __init__(self, *args, requestID=0, errorStatus=0,
                    errorIndex=0, variableBindings=None):
        self.requestID = requestID
        self.errorStatus = errorStatus
        self.errorIndex = errorIndex

        if variableBindings is None:
            self.variableBindings = VarBindList(*args)
        else:
            self.variableBindings = variableBindings

    def __len__(self):
        return 4

    def __repr__(self):
        args = []
        if self.requestID:
            args.append("requestID={}".format(self.requestID))

        if self.errorStatus:
            args.append("errorStatus={}".format(self.errorStatus))

        if self.errorIndex:
            args.append("errorIndex={}".format(self.errorIndex))

        args.append("variableBindings={}".format(repr(self.variableBindings)))
        return "{}({})".format(typename(self), ", ".join(args))

    def __str__(self, depth=0, tab="    "):
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            "{}{}:",
            "{}Request ID: {}",
            "{}Error Status: {}",
            "{}Error Index: {}",
            "{}Variable Bindings:",
            "{}"
        )).format(
            indent, typename(self),
            subindent, self.requestID,
            subindent, self.errorStatus,
            subindent, self.errorIndex,
            subindent, self.variableBindings.__str__(subindent + tab)
        )

    @property
    def objects(self):
        yield Integer(self.requestID)
        yield Integer(self.errorStatus)
        yield Integer(self.errorIndex)
        yield self.variableBindings

    @classmethod
    def deserialize(cls, data):
        requestID,   data = Integer.decode(data, leftovers=True)
        errorStatus, data = Integer.decode(data, leftovers=True)
        errorIndex,  data = Integer.decode(data, leftovers=True)
        variableBindings = VarBindList.decode(data)

        requestID = requestID.value
        errorStatus = errorStatus.value
        errorIndex = errorIndex.value

        try:
            errorStatus = cls.ErrorStatus(errorStatus)
        except ValueError as err:
            msg = "Invalid errorStatus: {}"
            raise ParseError(msg.format(errorStatus)) from err

        if errorStatus != cls.ErrorStatus.noError:
            if errorIndex < 0 or errorIndex > len(variableBindings):
                msg = "Error index {} not valid with {} variable bindings"
                raise ParseError(msg.format(errorIndex, len(variableBindings)))

        return cls(
            requestID=requestID,
            errorStatus=errorStatus,
            errorIndex=errorIndex,
            variableBindings=variableBindings,
        )

class BulkPDU(Constructed):
    def __init__(self, *args, requestID=0, nonRepeaters=0,
                    maxRepetitions=0, variableBindings=None):
        self.requestID = requestID
        self.nonRepeaters = nonRepeaters
        self.maxRepetitions = maxRepetitions

        if variableBindings is None:
            self.variableBindings = VarBindList(*args)
        else:
            self.variableBindings = variableBindings

    def __len__(self):
        return 4

    def __repr__(self):
        args = []
        if self.requestID:
            args.append("requestID={}".format(self.requestID))

        if self.nonRepeaters:
            args.append("nonRepeaters={}".format(self.nonRepeaters))

        if self.maxRepetitions:
            args.append("maxRepetitions={}".format(self.maxRepetitions))

        args.append("variableBindings={}".format(repr(self.variableBindings)))
        return "{}({})".format(typename(self), ", ".join(args))

    def __str__(self, depth=0, tab="    "):
        indent = tab * (depth + 1)
        return "\n".join((
            "{}:",
            "{}Request ID: {}",
            "{}Non-Repeaters: {}",
            "{}Max Repetitions: {}",
            "{}Variable Bindings:",
            "{}"
        )).format(
            typename(self),
            indent, self.requestID,
            indent, self.nonRepeaters,
            indent, self.maxRepetitions,
            indent, self.variableBindings.__str__(tab * (depth + 2))
        )

    @property
    def objects(self):
        yield Integer(self.requestID)
        yield Integer(self.nonRepeaters)
        yield Integer(self.maxRepetitions)
        yield self.variableBindings

    @classmethod
    def deserialize(cls, data):
        requestID,   data = Integer.decode(data, leftovers=True)
        nonRepeaters, data = Integer.decode(data, leftovers=True)
        maxRepetitions,  data = Integer.decode(data, leftovers=True)
        variableBindings = VarBindList.decode(data)

        if nonRepeaters.value < 0:
            raise ParseError("nonRepeaters may not be less than 0")
        elif maxRepetitions.value < 0:
            raise ParseError("maxRepetitions may not be less than 0")

        return cls(
            requestID=requestID.value,
            nonRepeaters=nonRepeaters.value,
            maxRepetitions=maxRepetitions.value,
            variableBindings=variableBindings,
        )

class Read:
    pass

class Write:
    pass

class Response:
    pass

class Notification:
    pass

class Internal:
    pass

class Confirmed:
    pass

class GetRequestPDU(PDU, Read, Confirmed):
    TYPE = Identifier(CLASS_CONTEXT_SPECIFIC, STRUCTURE_CONSTRUCTED, 0)

class GetNextRequestPDU(PDU, Read, Confirmed):
    TYPE = Identifier(CLASS_CONTEXT_SPECIFIC, STRUCTURE_CONSTRUCTED, 1)

class ResponsePDU(PDU, Response):
    TYPE = Identifier(CLASS_CONTEXT_SPECIFIC, STRUCTURE_CONSTRUCTED, 2)

class SetRequestPDU(PDU, Write, Confirmed):
    TYPE = Identifier(CLASS_CONTEXT_SPECIFIC, STRUCTURE_CONSTRUCTED, 3)

class GetBulkRequestPDU(BulkPDU, Read, Confirmed):
    TYPE = Identifier(CLASS_CONTEXT_SPECIFIC, STRUCTURE_CONSTRUCTED, 5)

class InformRequestPDU(PDU, Notification, Confirmed):
    TYPE = Identifier(CLASS_CONTEXT_SPECIFIC, STRUCTURE_CONSTRUCTED, 6)

class TrapPDU(PDU, Notification):
    TYPE = Identifier(CLASS_CONTEXT_SPECIFIC, STRUCTURE_CONSTRUCTED, 7)

class ReportPDU(PDU, Response, Internal):
    TYPE = Identifier(CLASS_CONTEXT_SPECIFIC, STRUCTURE_CONSTRUCTED, 8)

pduTypes = {
    cls.TYPE: cls for cls in (
        GetRequestPDU,
        GetNextRequestPDU,
        ResponsePDU,
        SetRequestPDU,
        GetBulkRequestPDU,
        InformRequestPDU,
        TrapPDU,
        ReportPDU,
    )
}



import cffi
from . import libcrypto

ffi = cffi.FFI()
ffi.cdef("""
#define AES_DECRYPT 0
#define AES_ENCRYPT 1
#define AES_MAXNR 14

struct aes_key_st {
    unsigned int rd_key[4 * (AES_MAXNR + 1)];
    int rounds;
};

typedef struct aes_key_st AES_KEY;

int AES_set_encrypt_key(const unsigned char *userKey, const int bits,
                        AES_KEY *key);
void AES_cfb128_encrypt(const unsigned char *in, unsigned char *out,
                        size_t length, const AES_KEY *key,
                        unsigned char *ivec, int *num, const int enc);
""")

ffi.set_source(
    "snmp.openssl.aes",
    "#include <openssl/aes.h>",
    libraries=[libcrypto],
)



__all__ = ["UdpTransport"]

import os
import select
import socket

from ..udp import UdpTransportBase
from snmp.utils import typename

RECV_SIZE = 65507

class UdpTransport(UdpTransportBase):
    def __init__(self, host="", port=0):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.r      = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.w      = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self.socket.setblocking(False)
        self.r     .setblocking(False)

        self.socket.bind((host, port))
        self.r.bind(("127.0.0.1", 0))
        self.w.bind(("127.0.0.1", 0))

    def close(self):
        self.w.close()
        self.r.close()
        self.socket.close()

    def listen(self, listener):
        sock = self.socket.fileno()
        stop = self.r.fileno()

        done = False
        while not done:
            ready = select.select([sock, stop], [], [])[0]
            for fd in ready:
                if fd == sock:
                    data, addr = self.socket.recvfrom(RECV_SIZE)
                    listener.hear(self, addr, data)
                elif fd == stop:
                    data, addr = self.r.recvfrom(1)
                    if data == b"\0" and addr == self.w.getsockname():
                        done = True

    def send(self, addr, packet):
        self.socket.sendto(packet, addr)

    def stop(self):
        self.w.sendto(b"\0", self.r.getsockname())



__all__ = ["SNMPv3UsmManager"]

import heapq
import threading
import weakref

from snmp.exception import *
from snmp.message import *
from snmp.pdu.v2 import *
from snmp.security import *
from snmp.security.levels import *
from snmp.types import *
from snmp.utils import *
from . import *

usmStatsNotInTimeWindowsInstance = OID.parse("1.3.6.1.6.3.15.1.1.2.0")
usmStatsUnknownEngineIDsInstance = OID.parse("1.3.6.1.6.3.15.1.1.4.0")

class State:
    def __init__(self, arg):
        if isinstance(arg, State):
            self._manager = arg._manager
        else:
            self._manager = weakref.ref(arg)

    @property
    def manager(self):
        manager = self._manager()
        assert manager is not None
        return manager

# User did not provide an engine ID, and
# has not yet attempted to send any requests
class Inactive(State):
    def onRequest(self, request):
        request.send()
        self.manager.state = WaitForDiscovery(self)

    def onReport(self, request, response):
        errmsg = "Received a report where no request should have been sent"
        raise SNMPLibraryBug(errmsg)

    def onResponse(self, request, response):
        errmsg = "Received a response where no request should have been sent"
        raise SNMPLibraryBug(errmsg)

# User provided an engine ID, but has not yet sent any requests
class Unsynchronized(State):
    def onRequest(self, request):
        request.send(self.manager.engineID)
        self.manager.state = Synchronizing(self)

    def onReport(self, request, response):
        errmsg = "Received a report where no request should have been sent"
        raise SNMPLibraryBug(errmsg)

    def onResponse(self, request, response):
        errmsg = "Received a response where no request should have been sent"
        raise SNMPLibraryBug(errmsg)

# User did not provide an engine ID, and the Manager has not
# yet received any communication from the remote engine.
class WaitForDiscovery(State):
    def onRequest(self, request):
        pass

    def onReport(self, request, response):
        engineID = response.securityEngineID

        try:
            self.manager.engineID = engineID
        except ValueError as err:
            print(err)
            return

        for entry in self.manager.requests:
            request = entry.request()
            if request is not None:
                request.send(engineID)

        self.manager.state = TrustEveryResponse(self)

    def onResponse(self, request, response):
        pass

# User provided and engineID and has sent at least one request, but the
# Manager has not yet received any communication from the remote engine.
class Synchronizing(State):
    def onRequest(self, request):
        if not request.securityLevel.auth:
            request.send(self.manager.engineID)

    def onReport(self, request, response):
        oid = response.data.pdu.variableBindings[0].name
        if oid == usmStatsUnknownEngineIDsInstance:
            try:
                request.send(response.securityEngineID)
            except ValueError:
                return
        elif response.securityLevel.auth:
            if response.securityEngineID != self.manager.engineID:
                self.manager.engineID = response.securityEngineID

            for entry in self.manager.requests:
                request = entry.request()
                if request is not None and request.securityLevel.auth:
                    request.send(self.manager.engineID)

            self.manager.state = RequireAuthentication(self)

    def onResponse(self, request, response):
        if response.securityLevel.auth:
            if response.securityEngineID != self.manager.engineID:
                self.manager.engineID = response.securityEngineID

        answered = request
        for entry in self.manager.requests:
            r = entry.request()
            if r is not None and r is not request and r.securityLevel.auth:
                r.send(self.manager.engineID)

        self.manager.state = RequireAuthentication(self)

class TrustEveryResponse(State):
    def onRequest(self, request):
        request.send(self.manager.engineID)

    def onReport(self, request, response):
        oid = response.data.pdu.variableBindings[0].name
        if oid == usmStatsUnknownEngineIDsInstance:
            try:
                request.send(response.securityEngineID)
            except ValueError:
                return

        if response.securityLevel.auth:
            if response.securityEngineID != self.manager.engineID:
                self.manager.engineID = response.securityEngineID
            self.manager.state = RequireAuthentication(self)

    def onResponse(self, request, response):
        if response.securityEngineID != self.manager.engineID:
            try:
                self.manager.engineID = response.securityEngineID
            except ValueError:
                return

        if response.securityLevel.auth:
            self.manager.state = RequireAuthentication(self)

class RequireAuthentication(State):
    def onRequest(self, request):
        request.send(self.manager.engineID)

    def onReport(self, request, response):
        oid = response.data.pdu.variableBindings[0].name
        if oid == usmStatsUnknownEngineIDsInstance:
            try:
                request.send(response.securityEngineID)
            except ValueError:
                return
        elif response.securityLevel.auth:
            if response.securityEngineID != self.manager.engineID:
                self.manager.engineID = response.securityEngineID

            if oid == usmStatsNotInTimeWindowsInstance:
                request.send(self.manager.engineID)

    def onResponse(self, request, response):
        if response.securityLevel.auth:
            if response.securityEngineID != self.manager.engineID:
                self.manager.engineID = response.securityEngineID

class Request:
    def __init__(self, pdu, manager, userName, securityLevel):
        self._engineID = None

        self.manager = manager
        self.pdu = pdu
        self.securityLevel = securityLevel
        self.userName = userName

        self.callback = None
        self.messages = set()

        self.event = threading.Event()
        self.expired = False
        self.fulfilled = False
        self.response = None

    def __del__(self):
        if self.callback is not None:
            self.close()

    @property
    def engineID(self):
        return self._engineID

    @engineID.setter
    def engineID(self, engineID):
        if engineID == self._engineID:
            return

        localEngine = self.manager.localEngine
        namespace =  self.manager.namespace

        if engineID is not None:
            if not localEngine.registerRemoteEngine(engineID, namespace):
                errmsg = "Failed to register engineID {} under namespace \"{}\""
                raise ValueError(errmsg.format(engineID, namespace))

        if self._engineID != None:
            localEngine.unregisterRemoteEngine(self._engineID, namespace)

        self._engineID = engineID

    def close(self):
        for message in self.messages:
            self.callback(message)

        self.callback = None
        self.messages.clear()

        self.engineID = None

    def addCallback(self, callback, msgID):
        if self.callback is None:
            self.callback = callback

        assert self.callback == callback
        self.messages.add(msgID)

    def push(self, response, depth=0):
        #if random.randint(1, 3) % 3 != 0:
        #    return

        if self.manager.processResponse(self, response):
            self.response = response
            self.event.set()

    def send(self, engineID=None):
        pdu = self.pdu
        user = self.userName
        securityLevel = self.securityLevel

        if engineID is None:
            if self.engineID is None:
                pdu = GetRequestPDU()
                engineID = b""
                user = b""
                securityLevel = noAuthNoPriv
            else:
                engineID = self.engineID
        else:
            self.engineID = engineID

        return self.manager.sendPdu(pdu, self, engineID, user, securityLevel)

    def wait(self):
        pdu = None
        while not self.expired:
            timeout = self.manager.refresh()
            #print(f"Waiting for {timeout} seconds")
            if self.event.wait(timeout=timeout):
                pdu = self.response.data.pdu
                self.fulfilled = True
                break

        self.close()
        if pdu is None:
            raise Timeout()
        else:
            return pdu

class SNMPv3UsmManager:
    def __init__(self, engine, locator, namespace,
            defaultUserName, defaultSecurityLevel,
            engineID=None, autowait=True):
        self._engineID = None

        self.autowait = autowait
        self.generator = NumberGenerator(32)
        self.localEngine = engine
        self.locator = locator
        self.namespace = namespace

        self.defaultUserName = defaultUserName
        self.defaultSecurityLevel = defaultSecurityLevel

        self.lock = threading.Lock()
        self.requests = []
        if engineID is None:
            self.state = Inactive(self)
        else:
            self.engineID = engineID
            self.state = Unsynchronized(self)

    def __del__(self):
        self.engineID = None

    @property
    def engineID(self):
        return self._engineID

    @engineID.setter
    def engineID(self, engineID):
        if engineID == self._engineID:
            return

        if engineID is not None:
            if not self.localEngine.registerRemoteEngine(engineID, self.namespace):
                errmsg = "Failed to register engineID {} under namespace \"{}\""
                raise ValueError(errmsg.format(engineID, self.namespace))

        if self._engineID != None:
            self.localEngine.unregisterRemoteEngine(self._engineID, self.namespace)

        self._engineID = engineID

    def refresh(self):
        while True:
            with self.lock:
                entry = self.requests[0]
                result = entry.refresh()

                while result is None:
                    heapq.heappop(self.requests)
                    entry = self.requests[0]
                    result = entry.refresh()

                #print(f"Result = {result}")

                if result < 0:
                    heapq.heapreplace(self.requests, entry)
                else:
                    return result

    def processResponse(self, request, response):
        with self.lock:
            if isinstance(response.data.pdu, ReportPDU):
                self.state.onReport(request, response)
                return False
            else:
                self.state.onResponse(request, response)
                return True

    def sendPdu(self, pdu, handle, engineID, user, securityLevel):
        return self.localEngine.dispatcher.sendPdu(
            self.locator.domain,
            self.locator.address,
            MessageProcessingModel.SNMPv3,
            pdu,
            handle,
            engineID,
            user,
            securityLevel=securityLevel,
            securityModel=SecurityModel.USM,
        )

    def sendRequest(self, pdu, securityLevel=None,
                    user=None, wait=None, **kwargs):
        if securityLevel is None:
            securityLevel = self.defaultSecurityLevel

        if user is None:
            user = self.defaultUserName

        if wait is None:
            wait = self.autowait

        request = Request(pdu, self, user, securityLevel, **kwargs)

        with self.lock:
            heapq.heappush(self.requests, Repeater(request))
            self.state.onRequest(request)

        if wait:
            return request.wait()
        else:
            return request

    def get(self, *oids, **kwargs):
        pdu = GetRequestPDU(*oids, requestID=next(self.generator))
        return self.sendRequest(pdu, **kwargs)

    def getBulk(self, *oids, nonRepeaters=0, maxRepetitions=0, **kwargs):
        pdu = GetBulkRequestPDU(
            *oids,
            requestID=next(self.generator),
            nonRepeaters=nonRepeaters,
            maxRepetitions=maxRepetitions,
        )

        return self.sendRequest(pdu, **kwargs)

    def getNext(self, *oids, **kwargs):
        pdu = GetNextRequestPDU(*oids, requestID=next(self.generator))
        return self.sendRequest(pdu, **kwargs)

    def set(self, *varbinds, **kwargs):
        varbinds = (VarBind(*varbind) for varbind in varbinds)
        pdu = SetRequestPDU(*varbinds, requestID=next(self.generator))
        return self.sendRequest(pdu, **kwargs)



import cffi
from . import libcrypto

ffi = cffi.FFI()
ffi.cdef("""
#define DES_DECRYPT 0
#define DES_ENCRYPT 1

typedef unsigned char DES_cblock[8];
typedef unsigned char const_DES_cblock[8];
typedef unsigned int DES_LONG;
typedef struct DES_ks {
    union {
        DES_cblock cblock;
        DES_LONG deslong[2];
    } ks[16];
} DES_key_schedule;

void DES_set_odd_parity(DES_cblock *key);
void DES_set_key_unchecked(const_DES_cblock *key, DES_key_schedule *schedule);
void DES_cbc_encrypt(const unsigned char *input, unsigned char *output,
                     long length, DES_key_schedule *schedule,
                     DES_cblock *ivec, int enc);
""")

ffi.set_source(
    "snmp.openssl.des",
    "#include <openssl/des.h>",
    libraries=[libcrypto],
)



__all__ = [
    "HmacMd5", "HmacSha", "HmacSha224", "HmacSha256", "HmacSha384", "HmacSha512"
]

import hashlib
import hmac

class AuthProtocol:
    def __init__(self, key):
        self.key = key

    @classmethod
    def localize(cls, secret, engineID):
        repeat, truncate = divmod(1 << 20, len(secret))

        context = cls.ALGORITHM()
        for i in range(repeat):
            context.update(secret)

        context.update(secret[:truncate])
        key = context.digest()

        context = cls.ALGORITHM()
        context.update(key)
        context.update(engineID)
        context.update(key)
        return context.digest()

    @property
    def msgAuthenticationParameters(self):
        return b'\0' * self.N

    def sign(self, data):
        context = hmac.new(self.key, digestmod=self.ALGORITHM)
        context.update(data)
        return context.digest()[:self.N]

class HmacMd5(AuthProtocol):
    ALGORITHM = hashlib.md5
    N = 12

class HmacSha(AuthProtocol):
    ALGORITHM = hashlib.sha1
    N = 12

class HmacSha224(AuthProtocol):
    ALGORITHM = hashlib.sha224
    N = 16

class HmacSha256(AuthProtocol):
    ALGORITHM = hashlib.sha256
    N = 24

class HmacSha384(AuthProtocol):
    ALGORITHM = hashlib.sha384
    N = 32

class HmacSha512(AuthProtocol):
    ALGORITHM = hashlib.sha512
    N = 48



__all__ = ["Aes128Cfb"]

import os
from snmp.openssl.aes import ffi, lib
from . import DecryptionError

class Aes128Cfb:
    BITS = 128
    INTSIZE = 4
    BYTEORDER = "big"

    BLOCKLEN = BITS // 8
    KEYLEN = BLOCKLEN
    SALTLEN = BLOCKLEN - (2 * INTSIZE)
    SALTWRAP = 1 << (8 * SALTLEN)

    def __init__(self, key):
        if len(key) < self.KEYLEN:
            errmsg = "key must be at least {} bytes long".format(self.KEYLEN)
            raise ValueError(errmsg)

        self.key = ffi.new("AES_KEY*")
        result = lib.AES_set_encrypt_key(key[:self.KEYLEN], self.BITS, self.key)
        assert result == 0

        self.salt = int.from_bytes(os.urandom(self.SALTLEN), self.BYTEORDER)

    def packIV(self, engineBoots, engineTime, salt):
        if len(salt) != self.SALTLEN:
            raise ValueError("Invalid salt")

        return b''.join((
            engineBoots.to_bytes(self.INTSIZE, self.BYTEORDER),
            engineTime .to_bytes(self.INTSIZE, self.BYTEORDER),
            salt
        ))

    def compute(self, data, iv, mode):
        num = ffi.new("int*", 0)
        iv = ffi.new("unsigned char[]", iv)
        output = ffi.new("unsigned char[]", len(data))
        lib.AES_cfb128_encrypt(data, output, len(data), self.key, iv, num, mode)
        return bytes(output)

    def decrypt(self, data, engineBoots, engineTime, salt):
        try:
            iv = self.packIV(engineBoots, engineTime, salt)
        except ValueError as err:
            raise DecryptionError(err) from err

        return self.compute(data, iv, lib.AES_DECRYPT);

    def encrypt(self, data, engineBoots, engineTime):
        self.salt = (self.salt + 1) % self.SALTWRAP
        salt = self.salt.to_bytes(self.SALTLEN, self.BYTEORDER)
        iv = self.packIV(engineBoots, engineTime, salt)
        return salt, self.compute(data, iv, lib.AES_ENCRYPT)



__all__ = [
    "INTEGER", "OCTET_STRING", "NULL", "OBJECT_IDENTIFIER", "SEQUENCE",
    "Asn1Encodable", "Integer", "OctetString", "Null", "OID",
    "Constructed", "Sequence",
]

import re
from snmp.ber import *
from snmp.exception import *
from snmp.utils import typename

INTEGER             = Identifier(CLASS_UNIVERSAL, STRUCTURE_PRIMITIVE, 2)
OCTET_STRING        = Identifier(CLASS_UNIVERSAL, STRUCTURE_PRIMITIVE, 4)
NULL                = Identifier(CLASS_UNIVERSAL, STRUCTURE_PRIMITIVE, 5)
OBJECT_IDENTIFIER   = Identifier(CLASS_UNIVERSAL, STRUCTURE_PRIMITIVE, 6)
SEQUENCE            = Identifier(CLASS_UNIVERSAL, STRUCTURE_CONSTRUCTED, 16)

class Asn1Encodable:
    def __eq__(a, b):
        return type(a) is type(b) and a.equals(b)

    @classmethod
    def decode(cls, data, leftovers=False, copy=True, **kwargs):
        result = decode(data, expected=cls.TYPE, leftovers=leftovers, copy=copy)

        if leftovers:
            encoding, leftovers = result
            return cls.deserialize(encoding, **kwargs), leftovers
        else:
            return cls.deserialize(result, **kwargs)

    def encode(self):
        return encode(self.TYPE, self.serialize())

    def equals(a, b):
        errmsg = "{} does not implement equals()"
        raise IncompleteChildClass(errmsg.format(typename(a, True)))

    def appendToOID(self, oid):
        errmsg = "{} does not implement appendToOID()"
        raise IncompleteChildClass(errmsg.format(typename(self, True)))

    @classmethod
    def decodeFromOID(cls, nums):
        errmsg = "{} does not implement decodeFromOID()"
        raise IncompleteChildClass(errmsg.format(typename(cls, True)))

    @classmethod
    def deserialize(cls, data):
        errmsg = "{} does not implement deserialize()"
        raise IncompleteChildClass(errmsg.format(typename(cls, True)))

    def serialize(self):
        errmsg = "{} does not implement serialize()"
        raise IncompleteChildClass(errmsg.format(typename(self, True)))

class Integer(Asn1Encodable):
    BYTEORDER = "big"
    SIGNED = True
    SIZE = 4
    TYPE = INTEGER

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f"{typename(self)}({self.value})"

    def equals(a, b):
        return a.value == b.value

    def appendToOID(self, oid):
        return oid.extend(self.value)

    @classmethod
    def decodeFromOID(cls, nums):
        value = next(nums)

        try:
            value.to_bytes(cls.SIZE, cls.BYTEORDER, signed=cls.SIGNED)
        except OverflowError as err:
            errmsg = f"{typename(cls)} value out of range: {value}"
            raise OID.IndexDecodeError(errmsg) from err

        return cls(value)

    @classmethod
    def deserialize(cls, data):
        for i in range(len(data) - cls.SIZE):
            if data[i] != 0:
                msg = f"Encoding too large for {typename(cls)}"
                raise ParseError(msg)

        return cls(int.from_bytes(data, cls.BYTEORDER, signed=cls.SIGNED))

    def serialize(self):
        try:
            encoding = self.value.to_bytes(
                self.SIZE,
                self.BYTEORDER,
                signed=self.SIGNED
            )
        except OverflowError as err:
            raise ValueError(err) from err

        if encoding[0]:
            return encoding

        for index in range(1, len(encoding)):
            if encoding[index]:
                if self.SIGNED and encoding[index] & 0x80:
                    index -= 1
                break

        return encoding[index:]

class OctetString(Asn1Encodable):
    TYPE = OCTET_STRING

    MIN_SIZE = 0
    MAX_SIZE = 0xffff
    INVALID_SIZES = ()

    def __init__(self, data=b""):
        self.data = data

    def __repr__(self):
        return f"{typename(self)}({self.data})"

    def equals(a, b):
        return a.data == b.data

    def appendToOID(self, oid):
        return oid.extend(len(self.data), *self.data)

    @classmethod
    def decodeFromOID(cls, nums):
        length = next(nums)
        data = bytearray(length)

        for i in range(length):
            byte = next(nums)

            try:
                data[i] = byte
            except ValueError as err:
                errmsg = "Sub-identifier does not fit within a single octet: {}"
                raise OID.IndexDecodeError(errmsg.format(byte)) from err

        return cls(bytes(data))

    @classmethod
    def deserialize(cls, data):
        if len(data) < cls.MIN_SIZE:
            msg = "Encoded {} may not be less than {} bytes long"
            raise ParseError(msg.format(typename(cls), cls.MIN_SIZE))
        elif len(data) > cls.MAX_SIZE:
            msg = "Encoded {} may not be more than {} bytes long"
            raise ParseError(msg.format(typename(cls), cls.MAX_SIZE))
        elif len(data) in cls.INVALID_SIZES:
            msg = "Encoded {} not permitted to be {} bytes long"
            raise ParseError(msg.format(typename(cls), len(data)))

        return cls.parse(data)

    @classmethod
    def parse(cls, data):
        return cls(data)

    def serialize(self):
        data = self.data
        if len(data) < self.MIN_SIZE:
            msg = "Encoded {} may not be less than {} bytes long"
            raise ValueError(msg.format(typename(self), self.MIN_SIZE))
        elif len(data) > self.MAX_SIZE:
            msg = "Encoded {} may not be more than {} bytes long"
            raise ValueError(msg.format(typename(self), self.MAX_SIZE))
        elif len(data) in self.INVALID_SIZES:
            msg = "Encoded {} not permitted to be {} bytes long"
            raise ValueError(msg.format(typename(self), len(data)))

        return data

class Null(Asn1Encodable):
    TYPE = NULL

    def __repr__(self):
        return f"{typename(self)}()"

    def equals(a, b):
        return isinstance(b, Null)

    def appendToOID(self, oid):
        return oid

    @classmethod
    def decodeFromOID(cls, nums):
        return cls()

    @classmethod
    def deserialize(cls, data):
        return cls()

    def serialize(self):
        return b''

class OID(Asn1Encodable):
    DOT = '.'
    MULT = 40
    MAXLEN = 128
    TYPE = OBJECT_IDENTIFIER

    class BadPrefix(IncomingMessageError):
        pass

    class IndexDecodeError(IncomingMessageError):
        pass

    def __init__(self, *nums):
        if len(nums) > self.MAXLEN:
            errmsg = "{} may not contain more than {} sub-identifiers"
            raise ValueError(errmsg.format(typename(self), self.MAXLEN))

        assert all(0 <= n < (1 << 32) for n in nums)
        self.nums = nums

    def __repr__(self):
        return f"{typename(self)}{self.nums}"

    def __str__(self):
        return self.DOT.join(str(n) for n in self.nums)

    def __getitem__(self, idx):
        return self.nums.__getitem__(idx)

    def __len__(self):
        return self.nums.__len__()

    def __lt__(a, b):
        return a.nums < b.nums

    def tryDecode(self, nums, cls):
        try:
            return cls.decodeFromOID(nums)
        except StopIteration as err:
            errmsg = f"Incomplete {typename(cls)} index"
            raise OID.IndexDecodeError(errmsg) from err

    FIRST = re.compile(r"^\.?(\d+|$)")
    REGEX = re.compile(r"\.(\d+)")

    @classmethod
    def parse(cls, oid):
        match = cls.FIRST.match(oid)

        if match is None:
            raise ValueError(f"Invalid OID string: {oid}")

        nums = []
        if match.group(1):
            while match is not None:
                nums.append(int(match.group(1)))
                index = match.end()
                match = cls.REGEX.match(oid, index)

            if index != len(oid):
                raise ValueError(f"Trailing characters in OID string: {oid}")

        try:
            if nums[0] > 2:
                errmsg = "{} may not begin with {}"
                raise ValueError(errmsg.format(typename(cls), nums[0]))

            if nums[1] >= cls.MULT:
                errmsg = "second number in {} must be less than {}"
                raise ValueError(errmsg.format(typename(cls), nums[1]))
        except IndexError as err:
            pass

        if any(n < 0 for n in nums):
            raise ValueError("\"{}\" contains a negative sub-identifier")
        elif any(n >= (1 << 32) for n in nums):
            errmsg = "OID \"{}\" contains a sub-identifier that is too large"
            raise ValueError(errmsg.format(oid))

        return cls(*nums)

    def extend(self, *nums):
        nums = self.nums + nums
        return type(self)(*nums)

    def appendIndex(self, *index):
        oid = self
        for obj in index:
            oid = obj.appendToOID(oid)

        return oid

    def extractIndex(self, prefix, *types):
        nums = iter(self.nums)
        for subid in prefix:
            try:
                match = (subid == next(nums))
            except StopIteration as err:
                errmsg = "\"{}\" is shorter than the given prefix \"{}\""
                raise self.BadPrefix(errmsg.format(self, prefix)) from err

            if not match:
                errmsg = "\"{}\" does not begin with \"{}\""
                raise self.BadPrefix(errmsg.format(self, prefix))

        index = None
        if len(types) == 1:
            index = self.tryDecode(nums, types[0])
        elif types:
            index = [None] * len(types)
            for i, cls in enumerate(types):
                index[i] = self.tryDecode(nums, cls)

            index = tuple(index)

        try:
            next(nums)
        except StopIteration:
            return index
        else:
            raise self.IndexDecodeError("Not all sub-identifiers were consumed")

    def equals(a, b):
        return a.nums == b.nums

    def appendToOID(self, oid):
        return oid.extend(len(self.nums), *self.nums)

    @classmethod
    def decodeFromOID(cls, nums):
        length = next(nums)
        subids = [0] * length

        for i in range(length):
            subids[i] = next(nums)

        return cls(*subids)

    @classmethod
    def deserialize(cls, data):
        data = iter(data)

        try:
            oid = list(divmod(next(data), cls.MULT))
        except StopIteration as err:
            raise ParseError(f"Empty {typename(cls)}") from err

        value = 0
        for byte in data:
            value |= byte & 0x7f
            if byte & 0x80:
                value <<= 7
                if value >= (1 << 32):
                    raise ParseError("Sub-identifier out of range")
            else:
                oid.append(value)
                value = 0

        if value:
            raise ParseError("OID ended unexpectedly")

        return cls(*oid)

    def serialize(self):
        def append(bytearr, num):
            if num < 0x80:
                bytearr.append(num)
            else:
                flag = 0
                tmp = bytearray()

                while num:
                    tmp.append((num & 0x7f) | flag)
                    flag = 0x80
                    num >>= 7

                tmp.reverse()
                bytearr.extend(tmp)

        try:
            first = self.nums[0]
        except IndexError:
            return b"\x00"

        try:
            second = self.nums[1]
        except IndexError:
            second = 0

        encoding = bytearray()
        append(encoding, first * self.MULT | second)
        for number in self.nums[2:]:
            append(encoding, number)

        return bytes(encoding)

class Constructed(Asn1Encodable):
    def equals(a, b):
        if len(a) == len(b):
            for left, right in zip(a.objects, b.objects):
                if left != right:
                    return False

        return True

    def __len__(self):
        errmsg = "{} does not implement __len__"
        raise IncompleteChildClass(errmsg.format(typename(self, True)))

    @property
    def objects(self):
        errmsg = "{} does not implement .objects"
        raise IncompleteChildClass(errmsg.format(typename(self, True)))

    def serialize(self):
        return b''.join([item.encode() for item in self.objects])

class Sequence(Constructed):
    TYPE = SEQUENCE



import importlib
import os
import os.path
import re
import sys

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

class build_openssl(build_ext):
    REGEX = re.compile(r"^snmp((?:\.[^.]+)+)$")

    def run(self):
        from cffi.recompiler import make_c_source
        sys.path.insert(0, os.path.dirname(__file__))

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        indices = []
        for index, ext in enumerate(self.extensions):
            match = self.REGEX.match(ext.name)
            if match is not None:
                ffi = importlib.import_module("snmp.cffi" + match.group(1)).ffi
                module, preamble, fileext = ffi._assigned_source[:3]
                outfile = os.path.join(self.build_temp, f"{module}{fileext}")
                make_c_source(ffi, module, preamble, outfile)
                ext.sources.append(outfile)
                indices.append(index)

        sys.path.pop(0)

        try:
            return super().run()
        except Exception:
            print("Failed to compile optional 'snmp.openssl' package")

        for index in reversed(indices):
            self.extensions.pop(index)

        return super().run()

sys.path.insert(0, os.path.dirname(__file__))

extensions = []
for name in ("aes", "des"):
    ffi = importlib.import_module(f"snmp.cffi.openssl.{name}").ffi
    module = ffi._assigned_source[0]
    kwargs = ffi._assigned_source[3]
    extensions.append(Extension(module, [], **kwargs))

sys.path.pop(0)

setup(
    cmdclass = {"build_ext": build_openssl},
    ext_modules=extensions,
    install_requires=["cffi>=1.0.0"],
    setup_requires=["cffi>=1.0.0"],
)



__all__ = [
    "Integer32", "IpAddress", "Counter32", "Gauge32",
    "Unsigned32", "TimeTicks", "Opaque", "Counter64",
    "zeroDotZero",
]

from socket import inet_aton, inet_ntoa
from snmp.ber import *
from snmp.types import *
from snmp.utils import typename

class Unsigned(Integer):
    SIGNED = False

class Integer32(Integer):
    pass

class IpAddress(OctetString):
    TYPE = Identifier(CLASS_APPLICATION, STRUCTURE_PRIMITIVE, 0)
    MIN_SIZE = 4
    MAX_SIZE = 4

    def __init__(self, addr):
        self.addr = addr

    def __repr__(self):
        return f"{typename(self)}({repr(self.addr)})"

    def equals(a, b):
        return a.data == b.data

    @property
    def data(self):
        try:
            return inet_aton(self.addr)
        except OSError as err:
            raise ValueError(f"Invalid IPv4 address: \"{self.addr}\"") from err

    @classmethod
    def parse(cls, data):
        return cls(inet_ntoa(data))

class Counter32(Unsigned):
    TYPE = Identifier(CLASS_APPLICATION, STRUCTURE_PRIMITIVE, 1)

class Unsigned32(Unsigned):
    TYPE = Identifier(CLASS_APPLICATION, STRUCTURE_PRIMITIVE, 2)

Gauge32 = Unsigned32

class TimeTicks(Unsigned):
    TYPE = Identifier(CLASS_APPLICATION, STRUCTURE_PRIMITIVE, 3)

class Opaque(OctetString):
    TYPE = Identifier(CLASS_APPLICATION, STRUCTURE_PRIMITIVE, 4)

class Counter64(Unsigned):
    SIZE = 8
    TYPE = Identifier(CLASS_APPLICATION, STRUCTURE_PRIMITIVE, 6)

zeroDotZero = OID(0, 0)



__all__ = ["UdpTransport"]

import importlib
from socket import AF_INET, inet_pton
from . import Transport, TransportDomain, package

class UdpTransportBase(Transport):
    DEFAULT_PORT = 161

    @staticmethod
    def normalizeAddress(address):
        if isinstance(address, tuple):
            ip, port = address
        else:
            ip = address
            port = UdpTransportBase.DEFAULT_PORT

        try:
            inet_pton(AF_INET, ip)
        except OSError as err:
            raise ValueError("Invalid IPv4 address: {}".format(ip)) from err

        if port <= 0 or port > 0xffff:
            errmsg = "Invalid UDP port number: {}"
            raise ValueError(errmsg.format(port)) from err

        return ip, port

module = importlib.import_module(".udp", package=package)
UdpTransport = module.UdpTransport
UdpTransport.DOMAIN = TransportDomain.UDP
